#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import re
import sys

import networkx

from usurper import pos


def sentences_iter(corpus, return_id=False):
    """Iterate over the sentences in a corpus.
    
    Args:
        corpus:

    """
    pattern = re.compile(r"^\d+\t")
    sentence = []
    sid = 0
    for line in corpus:
        if sys.version_info[:1] < (3,):
            line = line.decode("utf-8").rstrip("\n")
        else:
            line = line.rstrip("\n")
        if line == "":
            if len(sentence) > 0:
                sid += 1
                if return_id:
                    yield sentence, sid
                else:
                    yield sentence
            sentence = []
        else:
            m = re.search(pattern, line)
            if m:
                sentence.append(line.split("\t"))
            else:
                raise Exception("Line does not match: %s" % line)
    # in case the file does not end in an empty line
    if len(sentence) > 0:
        sid += 1
        if return_id:
            yield sentence, sid
        else:
            yield sentence


def create_nx_digraph_from_conll(conll, tagset, origid=None):
    """Return a networkx.DiGraph object of the CONLL representation,
    containing only the ID, FORM, CPOSTAG or POSTAG fields.
    
    Args:
        conll:
        origid:

    """
    dg = networkx.DiGraph()
    attributes = lambda l: {"form": l[1], "lower": l[1].lower(), "pos": l[pos_field]}    
    if tagset is not None:
        mapping = pos.get_pos_map(tagset)
        attributes = lambda l: {"form": l[1], "lower": l[1].lower(), "pos": l[pos_field], "upos": mapping[l[pos_field]]}
    if origid is not None:
        dg.graph["origid"] = origid
    pos_field = 4
    if conll[0][pos_field] == "_":
        pos_field = 3
    dg.add_nodes_from([(int(l[0]), attributes(l)) for l in conll])
    return dg


def export_to_conll_format(digraph):
    """Do actual formatting work.
    
    Args:
        digraph:

    """
    output = []
    for v in sorted(digraph.nodes()):
        if v == 0:
            continue
        word_id = v
        word = digraph.node[v]["form"]
        lemma = digraph.node[v].get("lemma", "_")
        cpostag = digraph.node[v].get("upos", "_")
        postag = digraph.node[v].get("pos", "_")
        feats = digraph.node[v].get("feats", "_")
        indeps = [(s, l["relation"]) for s, t, l in digraph.in_edges(v, data=True)]
        head, deprel = "_", "_"
        if len(indeps) == 1:
            head = indeps[0][0]
            deprel = indeps[0][1]
        else:
            logging.warning("Graph %s is not a tree: vertex %s has more than one incoming edge" % (digraph.graph["origid"], v))
        phead, pdeprel = "_", "_"
        output.append([str(word_id), word, lemma, cpostag, postag, feats, str(head), deprel, phead, pdeprel])
    return output



#!/usr/bin/python

import networkx
import operator
import re

def _words_iter(sentences, conll_format):
    """Iterator over words in sentences.
    
    Args:
        sentences: An iterator containing a lists of lists
            representing sentences (CWB format)

    Returns:
        An iterator

    """
    index = 0
    if conll_format:
        index = 1
    for sentence, sid in sentences:
        for token in sentence:
            yield token[index]


def get_top_n(sentences, n, k, conll_format):
    """A rudimentary implementation of TextRank (Mihalcea and Tarau, 2004)
    for keyword extraction. We do not use any filters or stop word
    lists, i.e. we expect to get the top n function words.
    
    Args:
        sentences: An iterator containing a lists of lists
            representing sentences (CWB format)
        n: number of keywords to extract
        k: window size
    
    Returns:
        The n highest ranked keywords, expected to be function words.

    """
    words = _words_iter(sentences, conll_format)
    window = ["" for _ in range(2 * k + 1)]
    g = networkx.Graph()
    for i, word in enumerate(words):
        window.pop(0)
        window.append(word.lower())
        if i < k:
            continue
        for j, w in enumerate(window):
            if j == k or w == "":
                continue
            g.add_edge(w, window[k])
    for i in range(k):
        window.pop(0)
        window.append("")
        for j, w in enumerate(window):
            if j == k or w == "":
                continue
            g.add_edge(w, window[k])
    pr = networkx.pagerank(g, alpha=1.0)
    pr = {key: value for key, value in pr.items() if not re.search(r"\W", key, re.U)}
    return [_[0] for _ in sorted(pr.items(), key=operator.itemgetter(1), reverse=True)[0:n]]



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import operator
import re

import networkx

from usurper.utils import tsv


def _add_edge(g, u, v):
    """Add edge to graph or increase edge weight."""
    if g.has_edge(u, v):
        g.edge[u][v]["weight"] += 1
    else:
        g.add_edge(u, v, {"weight": 1})
    return g


def _add_edges(s, function_words):
    """Create vertices and edges."""
    s_len = s.__len__()
    fw_set = set(["ADP", "CONJ", "DET", "PRON", "PRT"])
    for i in range(s_len):
        i += 1
        for j in range(s_len):
            j += 1
            if i == j:
                continue
            # short edges – edges between neighboring words
            if j == i - 1 or j == i + 1:
                s = _add_edge(s, i, j)
                # function words – edges from neighboring words to
                # function words
                # if "upos" in s.node[i]:
                #     if s.node[i]["upos"] in fw_set:
                #         s = _add_edge(s, j, i)
                # else:
                if s.node[i]["lower"] in function_words:
                    s = _add_edge(s, j, i)
            # morphological inequality – edges between words with
            # different pre- or suffixes
            if "upos" in s.node[i]:
                if s.node[i]["upos"] != s.node[j]["upos"] or s.node[i]["lower"][:2] != s.node[j]["lower"][:2] or s.node[i]["lower"][-3:] != s.node[j]["lower"][-3:]:
                    s = _add_edge(s, i, j)
            else:
                if s.node[i]["lower"][:2] != s.node[j]["lower"][:2] or s.node[i]["lower"][-3:] != s.node[j]["lower"][-3:]:
                    s = _add_edge(s, i, j)
            # verb edges
            if "upos" in s.node[i]:
                if s.node[i]["upos"] == "VERB":
                    s = _add_edge(s, j, i)
                    if s.node[j]["upos"] == "NOUN":
                        s = _add_edge(s, i, j)
    return s


def _rank_vertices(sid, digraph):
    """Rank vertices using PageRank."""
    pr = None
    try:
        pr = networkx.pagerank(digraph, alpha=1.0, tol=1e-05, weight="weight")
    except networkx.exception.NetworkXError:
        if sid is not None:
            logging.warning("PageRank did not converge on sentence %s" % sid)
        else:
            logging.warning("PageRank did not converge")
        pr = {v: -1.0 * v for v in digraph.nodes()}
    return pr


def _construct_parse_tree(s, pageranks, no_rules):
    """Construct dependency parse tree from PageRanks."""
    # potential governors
    # Paper:
    dep_rules = {"VERB": set(["VERB"]), "NOUN": set(["VERB", "NOUN", "ADP"]), "ADJ": set(["NOUN", "VERB"]), "ADV": set(["VERB", "ADP"]), "ADP": set(["VERB"]), "CONJ": set(["VERB"]), "DET": set(["NOUN", "VERB"]), "NUM": set(["NOUN", "VERB"])}
    # Paper + VERB -> PRON
    # dep_rules = {"VERB": set(["VERB"]), "NOUN": set(["VERB", "NOUN", "ADP"]), "ADJ": set(["NOUN", "VERB"]), "ADV": set(["VERB", "ADP"]), "ADP": set(["VERB"]), "CONJ": set(["VERB"]), "DET": set(["NOUN", "VERB"]), "NUM": set(["NOUN", "VERB"]), "PRON": set(["VERB"])}
    g = networkx.DiGraph()
    g.add_nodes_from(s.nodes(data=True))
    pr = sorted(pageranks.items(), key=operator.itemgetter(1), reverse=True)
    pr_dict = {v: r for v, r in pr}
    heads = set()
    # Most central verb as root, if possible
    if "upos" in g.node[pr[0][0]] and not no_rules:
        root_candidates = [v for v, r in pr if g.node[v]["upos"] == "VERB"]
        if len(root_candidates) > 0:
            root = root_candidates[0]
            g.add_edge(0, root, {"relation": "ROOT"})
            heads = set([root])
    for v, r in pr:
        if len(heads) == 0:
            g.add_edge(0, v, {"relation": "ROOT"})
            heads.add(v)
            continue
        if g.has_edge(0, v):
            continue
        # TODO: penalize punctuation as head
        parent = None
        parent_candidates = []
        if "upos" in g.node[v] and not no_rules:
            parent_upos = dep_rules.get(g.node[v]["upos"], set())
            parent_candidates = [u for u in heads if g.node[u]["upos"] in parent_upos]
        if len(parent_candidates) > 0:
            # if g.node[v]["upos"] == "DET":
            #     right = [pc for pc in parent_candidates if pc > v]
            #     if len(right) > 0:
            #         parent = min(right, key=lambda u: (abs(u - v), pr_dict[u] * -1.0))
            #     else:
            #         parent = min(parent_candidates, key=lambda u: (abs(u - v), pr_dict[u] * -1.0))
            # else:
            parent = min(parent_candidates, key=lambda u: (abs(u - v), pr_dict[u] * -1.0))
        else:
            parent = min(heads, key=lambda u: (abs(u - v), pr_dict[u] * -1.0))
        heads.add(v)
        g.add_edge(parent, v, {"relation": "DEP"})
    return g


def parse_sentence_graph(args):
    """Parse sentence graph using the algorithm by Søgaard (2012)."""
    (s, sid), function_words, no_rules = args
    s = _add_edges(s, function_words)
    pr = _rank_vertices(sid, s)
    p = _construct_parse_tree(s, pr, no_rules)
    return p


def parse_sentence(tokens, function_words, no_rules, tags=[], tagset=None):
    """Parse sentence using the algorithm by Søgaard (2012).
    
    Args:
        tokens: list of tokens
        function_words: set of function words
        no_rules: boolean; true if universal dependency rules should
            not be used
        tags: list of tags, if available; the nth element of tags
            should be the part-of-speech tag associated with the nth
            element of tokens
        tagset: string identifying one of the supported tagsets
    
    Returns:
        A networkx DiGraph representing the dependency structure.

    """
    sentence = None
    if len(tokens) == len(tags):
        sentence = list(zip(tokens, tags))
    else:
        if len(tags) > 0:
            logging.warning("len(tokens) != len(tags) -- will not use part-of-speech tags")
        sentence = list(zip(tokens))
    s = tsv.create_nx_digraph_from_tsv(sentence, tagset)
    return parse_sentence_graph(((s, None), function_words, no_rules))



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import re

import networkx

from usurper import pos


def sentences_iter(corpus, return_id=False):
    """Iterate over the sentences in a corpus.
    
    Args:
        corpus:

    """
    pattern = re.compile(r"^\S+")
    sentence = []
    sid = 0
    for line in corpus:
        line = line.rstrip("\n")
        if line == "":
            if len(sentence) > 0:
                sid += 1
                if return_id:
                    yield sentence, sid
                else:
                    yield sentence
            sentence = []
        else:
            m = re.search(pattern, line)
            if m:
                sentence.append(line.split("\t"))
            else:
                raise Exception("Line does not match: %s" % line)
    # in case the file does not end in an empty line
    if len(sentence) > 0:
        sid += 1
        if return_id:
            yield sentence, sid
        else:
            yield sentence


def create_nx_digraph_from_tsv(tsv, tagset, origid=None):
    """Return a networkx.DiGraph object of the TSV representation.
    
    Args:
        tsv:
        origid:

    """
    dg = networkx.DiGraph()
    attributes = lambda l: {"form": l[0], "lower": l[0].lower()}
    if len(tsv[0]) > 1:
        attributes = lambda l: {"form": l[0], "lower": l[0].lower(), "pos": l[1]}
        if tagset is not None:
            mapping = pos.get_pos_map(tagset)
            attributes = lambda l: {"form": l[0], "lower": l[0].lower(), "pos": l[1], "upos": mapping[l[1]]}
    if origid is not None:
        dg.graph["origid"] = origid
    dg.add_nodes_from([(i + 1, attributes(l)) for i, l in enumerate(tsv)])
    return dg



from setuptools import setup

setup(
    name='Usurper',
    version='0.9.1',
    author='Thomas Proisl',
    author_email='thopro@posteo.de',
    packages=[
        'usurper',
        'usurper.utils',
    ],
    scripts=[
        'bin/usrpr',
    ],
    package_data={
        'usurper': ["data/ar-padt.map", "data/bg-btb.map", "data/ca-cat3lb.map", "data/cs-pdt.map", "data/da-ddt.map", "data/de-negra.map", "data/de-tiger.map", "data/el-gdt.map", "data/en-brown.map", "data/en-ptb.map", "data/en-tweet.map", "data/es-cast3lb.map", "data/eu-eus3lb.map", "data/fi-tdt.map", "data/fr-paris.map", "data/hu-szeged.map", "data/it-isst.map", "data/iw-mila.map", "data/ja-kyoto.map", "data/ja-verbmobil.map", "data/ko-sejong.map", "data/nl-alpino.map", "data/pl-ipipan.map", "data/pt-bosque.map", "data/ru-rnc.map", "data/sl-sdt.map", "data/sv-talbanken.map", "data/tu-metusbanci.map", "data/zh-ctb6.map", "data/zh-sinica.map"]
    },
    url='http://pypi.python.org/pypi/Usurper/',
    license='GNU General Public License v3 or later (GPLv3+)',
    description='An unsupervised dependency parser.',
    long_description=open('README.txt').read(),
    install_requires=[
        "networkx",
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Text Processing :: Linguistic',
    ],
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import sys


def get_pos_map(tagset):
    """Return part-of-speech map for given tag set. See
    https://code.google.com/p/universal-pos-tags/

    """
    usurper_dir = os.path.dirname(os.path.realpath(__file__))
    # Usurper_dir = os.path.abspath(os.path.join(usurper_dir, os.pardir))
    # map_path = os.path.join(Usurper_dir, "data", "universal-pos-tags", tagset + ".map")
    map_path = os.path.join(usurper_dir, "data", tagset + ".map")
    mapping = {}
    with open(map_path) as fh:
        for line in fh:
            if sys.version_info[:1] < (3,):
                line = line.decode("utf-8").rstrip()
            else:
                line = line.rstrip()
            t = line.split("\t")
            mapping[t[0]] = t[1]
    return mapping
